from linked_list import LinkedListKVstore


class HashMap(object):

    _item_count_per_bucket = 10

    def __init__(self, bucket_count=3):
        self.length = 0
        self.bucket_count = bucket_count 
        self.arr = [LinkedListKVstore() for _ in range(self.bucket_count)]

    def __len__(self):
        return self.length

    def _hash(self, k):
        """
        >>> h = HashMap()
        >>> h._hash('a')
        1
        >>> h._hash('b')
        2
        >>> h._hash('c')
        0
        """
        return sum(ord(i) for i in k) % self.bucket_count

    def __contains__(self, k):
        """
        >>> h = HashMap()
        >>> h['a'] = 1
        >>> 'a' in h
        True
        >>> 'b' in h
        False
        """
        h = self._hash(k)
        bucket = self.arr[h]
        return k in bucket

    def __setitem__(self, k, v):
        """
        >>> h = HashMap()
        >>> h['a'] = 1
        >>> for i in range(40):
        ...     h[str(i)] = i
        ...
        >>> h.bucket_count
        7
        """
        if self.bucket_count * self._item_count_per_bucket <= self.length:
            new_bucket_count = self.bucket_count * 2 + 1
            new_hash_map = HashMap(new_bucket_count) 
            new_hash_map[k] = v
            for _k, _v in self:
                new_hash_map[_k] = _v
            self.length = new_hash_map.length
            self.bucket_count = new_hash_map.bucket_count
            self.arr = new_hash_map.arr
        else:
            h = self._hash(k)
            bucket = self.arr[h]
            if k not in bucket:
                self.length += 1
            bucket[k] = v

    def __getitem__(self, k):
        """
        >>> h = HashMap()
        >>> h['a'] = 1
        >>> h['b'] = 2
        >>> h['a']
        1
        >>> h['b']
        2
        """
        h = self._hash(k)
        bucket = self.arr[h]
        if k in bucket:
            return bucket[k]

    def __delitem__(self, k):
        """
        >>> h = HashMap()
        >>> h['a'] = 1
        >>> h['b'] = 2
        >>> len(h)
        2
        >>> del h['a']
        >>> 'a' in h
        False
        >>> 'b' in h
        True
        >>> len(h)
        1
        >>> for i in range(40):
        ...     h[str(i)] = i
        ...
        >>> h.bucket_count
        7
        >>> for i in range(40):
        ...     del h[str(i)]
        ...
        >>> h.bucket_count
        3
        """
        if self.bucket_count > 3 and self.length <= int(self.bucket_count / 2):
            new_bucket_count = int((self.bucket_count - 1) / 2)
            new_hash_map = HashMap(new_bucket_count)
            for _k, _v in self:
                if _k == k:
                    continue
                new_hash_map[_k] = _v
            self.length = new_hash_map.length
            self.bucket_count = new_hash_map.bucket_count
            self.arr = new_hash_map.arr
        else:
            h = self._hash(k)
            bucket = self.arr[h]
            if k in bucket:
                del bucket[k]
                self.length -= 1

    def __iter__(self):
        """
        >>> h = HashMap()
        >>> h['a'] = 1
        >>> h['b'] = 2
        >>> h['c'] = 3
        >>> for k, v in h:
        ...     print(k, v)
        c 3
        a 1
        b 2
        """
        for bucket in self.arr:
            for k, v in bucket:
                yield k, v 

    def __repr__(self):
        """
        >>> h = HashMap()
        >>> h['a'] = 1
        >>> h['b'] = 2
        >>> h['c'] = 3
        >>> h
        {'c': 3, 'a': 1, 'b': 2}
        """
        lst = []
        for bucket in self.arr:
            for k, v in bucket:
                lst.append('{}: {}'.format(repr(k), v))
        return '{' + ', '.join(lst) + '}'
    
if __name__ == '__main__':
    import doctest
    doctest.testmod()



class LinkedListNode(object):

    def __init__(self, k, v):
        self.k = k
        self.v = v
        self.next = None

    def __repr__(self):
        """
        >>> n = LinkedListNode('a', 1)
        >>> n
        Node('a', 1)
        """
        return 'Node({}, {})'.format(repr(self.k), self.v)


class LinkedListKVstore(object):
    
    def __init__(self):
        self.length = 0
        self.head = LinkedListNode(None, None)

    def __len__(self):
        return self.length

    def __setitem__(self, k, v):
        """
        >>> link = LinkedListKVstore()
        >>> link['a'] = 1
        >>> link['b'] = 2
        >>> len(link)
        2
        >>> link['a']
        1
        >>> link['b'] 
        2
        >>> link['a'] = 100
        >>> link['a']
        100
        """
        h = self._get_node(k)
        if h is not None:
            h.v = v
        else:
            n = LinkedListNode(k, v)
            n.next = self.head.next
            self.head.next = n
            self.length += 1

    def _get_node(self, k):
        h = self.head.next
        while h:
            if h.k == k:
                return h 
            h = h.next

    def __getitem__(self, k):
        """
        >>> link = LinkedListKVstore()
        >>> link['a'] = 1
        >>> link['b'] = 2
        >>> len(link)
        2
        >>> link['a']
        1
        >>> link['b'] 
        2
        """
        h = self._get_node(k)
        if h is not None:
            return h.v

    def __delitem__(self, k):
        """
        >>> link = LinkedListKVstore()
        >>> link['a'] = 1
        >>> link['b'] = 2
        >>> link['c'] = 3
        >>> len(link)
        3
        >>> del link['b']
        >>> 'b' in link
        False
        >>> len(link)
        2
        """
        h = self.head
        while h.next:
            if h.next.k == k:
                h.next = h.next.next
                self.length -= 1
                return
            h = h.next

    def __repr__(self):
        """
        >>> link = LinkedListKVstore()
        >>> link['a'] = 1
        >>> link['b'] = 2
        >>> link['c'] = 3
        >>> link
        {'c': 3, 'b': 2, 'a': 1}
        """
        k_v = ['{}: {}'.format(repr(k), v) for k, v in self]
        return '{' + ', '.join(k_v) + '}'

    def __iter__(self):
        """
        >>> link = LinkedListKVstore()
        >>> link['a'] = 1
        >>> link['b'] = 2
        >>> link['c'] = 3
        >>> for k, v in link:
        ...     print(k, v)
        c 3
        b 2
        a 1
        """
        h = self.head.next
        while h:
            yield h.k, h.v
            h = h.next

    def __contains__(self, k):
        """
        >>> link = LinkedListKVstore()
        >>> link['a'] = 1
        >>> link['b'] = 2
        >>> 'a' in link
        True
        >>> 'c' in link
        False
        """
        h = self._get_node(k)
        return h is not None


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import setuptools


setuptools.setup(
    name='xy-libcollection',
    version='0.0.7',
    author='xueyan',
    author_email='xueyan0407@outlook.com',
    description="xueyan's collection libs",
    long_description='#A long desc',
    long_descripthin_content_type='text/markdown',
    url='https://github.com/qimiaoxue/libcollection',
    packages=setuptools.find_packages(),
    classifiers=[],
)
